
from collections import defaultdict
import functools
import inspect
import weakref

from .helpers import require, Config, keep_if_not
from .mixins import LockMixin, LogMixin, StorageMixin, BaseLogMixin
from .stats import RunningStats


class NamedMethod:

    _func = None

    @property
    def name(self):
        return self._func.__name__

    @property
    def signature(self):
        return inspect.signature(self._func)

    @property
    def parameters(self):
        return tuple(self.signature.parameters.keys())

    def check_signature(self, func):
        pass

    def __set_name__(self, owner, name):
        pass

    def __call__(self, func):
        self.check_signature(func)
        self._func = func

        class NewCallable:

            def __get__(selfie, instance, owner=None):
                if instance is None:
                    return self

                func = functools.partial(self.call, instance)
                func.__wrapped__ = self._func
                return func

            def __name__(self):
                return func.__name__

            def __set_name__(selfie, owner, name):
                self.__set_name__(owner, name)

            def __getattr__(selfie, item):
                return getattr(self, item)

            def __call__(selfie, instance, *args, **kwargs):
                return self._func(instance, *args, **kwargs)

        obj = NewCallable()

        return obj

    def __newcall__(self, instance, *args, **kwargs):
        return self.call(instance, *args, **kwargs)

    def call(self, instance, *args, **kwargs):
        return self._func(instance, *args, **kwargs)

    def raw_call(self, instance, *args, **kwargs):
        return self._func(instance, *args, **kwargs)


class LockMethod(NamedMethod):

    def __set_name__(self, owner, name):
        require(self, owner, name, LockMixin)

        super().__set_name__(owner, name)

    def call(self, instance, *args, **kwargs):
        with instance.lock:
            return super().call(instance, *args, **kwargs)


class LogMethod(NamedMethod):

    log_values = True

    def __init__(self, *args, **kwargs):
        self.log_values = kwargs.pop('log_values', True)
        super().__init__(*args, **kwargs)

    def _to_log(self, instance, value):
        if self.log_values is True:
            return value

        elif callable(self.log_values):
            try:
                return self.log_values(value)
            except Exception as e:
                instance.log_error('Could not convert value to log in %s, logging type: e', self.name, e)
                return type(value)

        return type(value)

    def _args_kwargs_to_log(self, instance, args, kwargs):
        if self.log_values is True:
            return args, kwargs

        elif callable(self.log_values):
            try:
                return tuple(self.log_values(arg) for arg in args), {k: self.log_values(v) for k, v in kwargs.items()}
            except Exception as e:
                instance.log_error('Could not convert value to log in %s, logging type: e', self.name, e)

        return tuple(type(arg) for arg in args), {k: type(v) for k, v in kwargs.items()}

    def __set_name__(self, owner, name):
        require(self, owner, name, LogMixin)

        super().__set_name__(owner, name)

    def call(self, instance, *args, **kwargs):
        if args or kwargs:
            _args, _kwargs = self._args_kwargs_to_log(instance, args, kwargs)
            instance.log_info('Calling %s with (%s, %s))', self.name, _args, _kwargs)
        else:
            instance.log_info('Calling %s', self.name)

        try:
            out = super().call(instance, *args, **kwargs)
            instance.log_info('%s returned %s', self.name, self._to_log(instance, out))
            return out
        except Exception as e:
            instance.log_error('While calling %s: %s', self.name, e)
            raise e


class StorageMethod(NamedMethod):
    """A property that can store and retrieve information in the instance
    to which is attached.

    Methods and descriptors are class attributes and therefore any attempt to
    naively modify one of their attributes for a single instance of the parent
    class will propagate to all instances. This property overcomes this problem
    by storing information at the instance level.

    The information is stored in uniquely a specified namespace defined by
    the derived class. Inside that storage, another namespace is specified
    using the property name.

    Derived class should use the dynamically created _store_get and _store_set
    to retrieve and store information.

    Derived classes must override the following variables:

    _storage_ns : str
        Defines a unique namespace under which the information of the
        derived class is stored.

    _storage_ns_init : callable
        Called upon initialization of the storage to initialize the
        specific storage of the namespace.


    Requires that the owner class inherits StorageMixin.
    """

    # Stores namespace to StorageProperty subclass
    # It cannot be dunder because it is accessed by __init_subclass__
    _storage_sub_ns_cls = weakref.WeakValueDictionary()

    _storage_ns = ''
    _storage_ns_init = None

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

        ns = cls._storage_ns
        if not ns:
            raise ValueError('Class %s must specify a storage namespace '
                             ' as required by StorageMethod' % cls)

        if ns in cls._storage_sub_ns_cls:
            if not issubclass(cls, cls._storage_sub_ns_cls[ns]):
                raise ValueError('Class %r storage namespace (%s) collides with '
                                 'class %r' % (cls, ns, cls._storage_sub_ns_cls[ns]))
        else:
            if cls._storage_ns_init is None:
                raise ValueError('Class %s must specify a storage initializer '
                                 ' as required by StorageMethod' % cls)

            cls._storage_sub_ns_cls[ns] = cls

            # Create partial versions of _store_get and _store_get
            #  with the corresponding namespace
            #  and store it in the specific subclass.
            cls._store_get = functools.partial(cls._ns_store_get, namespace=cls._storage_ns)
            cls._store_set = functools.partial(cls._ns_store_set, namespace=cls._storage_ns)
            cls._store_del = functools.partial(cls._ns_store_del, namespace=cls._storage_ns)

    def __set_name__(self, owner, name):
        require(self, owner, name, StorageMixin)

        super().__set_name__(owner, name)

    def _store_get(self, instance):
        return self._ns_store_get(instance, self._storage_ns)

    def _store_set(self, instance, value):
        return self._ns_store_set(instance, value, self._storage_ns)

    def _store_del(self, instance):
        return self._ns_store_del(instance, self._storage_ns)

    def _ns_store_get(self, instance, namespace):
        sto = instance.storage

        if namespace not in sto:
            cls = self._storage_sub_ns_cls[namespace]
            sto[namespace] = cls._storage_ns_init(instance)

        return sto[namespace][self.name]

    def _ns_store_set(self, instance, value, namespace):
        sto = instance.storage

        if namespace not in sto:
            cls = self._storage_sub_ns_cls[namespace]
            sto[namespace] = cls._storage_ns_init(instance)

        sto[namespace][self.name] = value

    def _ns_store_del(self, instance, namespace):
        sto = instance.storage

        if namespace not in sto:
            cls = self._storage_sub_ns_cls[namespace]
            sto[namespace] = cls._storage_ns_init(instance)

        del sto[namespace][self.name]


class StatsMethod(StorageMethod):
    """A property that keep stats on get and set calls.

    Stats can be retrieved with the `stat` methods and
    the following keys:
    - call
    - failed_call

    The following statistics are provided in a namedtuple
    last : float
        most recent duration (seconds).
    count : int
        number of operations.
    mean : float
        average duration per operation (seconds).
    std : float
        standard deviation of the duration (seconds).
    min : float
        shortest duration (seconds).
    max : float
        longest duration (seconds).


    Requires that the owner class inherits StorageMixin.
    """

    _storage_ns = 'statsm'
    _storage_ns_init = lambda _: defaultdict(RunningStats)

    def __set_name__(self, owner, name):
        require(self, owner, name, LogMixin)
        super().__set_name__(owner, name)

    def call(self, instance, *args, **kwargs):
        with StatsMethod._store_get(self, instance).time('call'):
            return super().call(instance, *args, **kwargs)

    def stats(self, instance, key):
        return StatsMethod._store_get(self, instance).stats(key)


class InstanceConfigurableMethod(StorageMethod):

    _storage_ns = 'iconfigm'
    _storage_ns_init = lambda _: defaultdict(dict)

    _config_keys = None
    _config = None
    _config_unset_value = None

    def __init__(self, *args, **kwargs):
        self._config = {}
        for k in self._config_keys:
            if k in kwargs:
                self._config[k] = kwargs.pop(k)

        super().__init__()

    def config_get(self, instance, key):

        if instance is None:
            return self._config.get(key, self._config_unset_value)
        try:
            return InstanceConfigurableMethod._store_get(self, instance)[key]
        except KeyError:
            return self._config.get(key, None)

    def config_set(self, instance, key, value):
        if instance is None:
            self._config[key] = value
        else:
            InstanceConfigurableMethod._store_get(self, instance)[key] = value

        self.on_config_set(instance, key, value)

    def config_iter(self, instance):
        for key in self._config.keys():
            yield key, self.config_get(instance, key)

    def on_config_set(self, instance, key, value):
        pass


class TransformMethod(InstanceConfigurableMethod):

    _transformations = None
    _storage_ns = 'transformationsm'
    _storage_ns_init = lambda _: defaultdict(dict)

    params = Config()

    def __set_name__(self, owner, name):
        require(self, owner, name, StorageMixin, BaseLogMixin)

        super().__set_name__(owner, name)

        sig = self.signature
        param_names = self.parameters
        if self.params:
            if not isinstance(self.params, dict):
                if len(sig.parameters) > 2:
                    raise ValueError('The syntax for methods with multiple arguments '
                                     'in the constructor has been deprecated. Use the '
                                     'cleaner "param" syntax in %s', name)

                last = param_names[-1]
                self.params = {last: self.params}
        else:
            self.params = {}
            try:
                self.params = dict(self._func.__transform_params__)
                for k in self.__transform_params__.keys():
                    if k not in param_names:
                        raise ValueError('%s is not an argument name of %s', k, name)
                del self._func.__transform_params__
            except AttributeError:
                pass

        assert isinstance(self.params, dict)

    def check_signature(self, func):
        super().check_signature(func)

        if not self.params:
            return

        names = tuple(inspect.signature(func).parameters.keys())

        p = self.params.pop(None, None)

        if not p:
            return

        for name in names[1:]:
            if name not in self.params:
                self.params[name] = p

    def call(self, instance, *args, **kwargs):

        sig = self.signature
        ba = sig.bind(instance, *args, **kwargs)

        t_arg = self.params_iget(instance)
        if t_arg:
            r_arg = t_arg.pop('<ret>', None)
        else:
            r_arg = None
        new_ba = {}

        noop = lambda x: x

        if t_arg:
            
            for param in sig.parameters.values():
                if param.kind != param.POSITIONAL_OR_KEYWORD:
                    raise ValueError('Only named POSITIONAL_OR_KEYWORD arguments are currently '
                                     'supported by transformations.')

            for k, v in ba.arguments.items():
                new_ba[k] = t_arg.get(k, noop)(v)

            ba = sig.bind(**new_ba)
            instance.log_info('<T> Calling %s with (%s, %s)', self.name, ba.args, ba.kwargs)
            try:
                out = super().call(*ba.args, **ba.kwargs)
            except Exception as e:
                instance.log_error('While pre-processing (%s, %s) for %s: %s', ba.args, ba.kwargs, self.name, e)
                raise e
        else:
            out = super().call(*ba.args, **ba.kwargs)

        if r_arg:
            try:
                out = r_arg(out)
            except Exception as e:
                instance.log_error('While post-processing %s for %s: %s', out, self.name, e)
                raise e

        return out

    @classmethod
    def param(cls, names, func):
        """Add modifiers to a specific parameter.

        See Action for more information.
        """
        def decorator(f):
            _param_memo(f, names, func)
            return f
        return decorator

    @classmethod
    def ret(cls, func):
        """Add modifiers to the return value.

        See Action for more information.
        """
        def decorator(f):
            _param_memo(f, '<ret>', func)
            return f
        return decorator


def _param_memo(obj, name, func):
    """Helper function to allow composition of modifiers in a single action.
    """
    if not isinstance(name, str):
        for n in name:
            _param_memo(obj, n, func)
        return

    if not hasattr(obj, '__transform_params__'):
        obj.__transform_params__ = []
    obj.__transform_params__.append((name, func))




from collections import defaultdict
import functools as ft
import inspect
import weakref

from .helpers import missingdict, require, DictPropertyNameKey, Config
from .stats import RunningStats
from .mixins import StorageMixin, BaseLogMixin, LockMixin, CacheMixin, ObservableMixin


class NamedProperty:
    """A property that takes the name of the class attribute to which it is assigned.

    The name must be convertible to string.
    """

    name = ''
    kwargs = None

    def __init__(self, fget=None, fset=None, fdel=None, doc=None):
        self.fget = fget
        self.fset = fset
        self.fdel = fdel
        if doc is None and fget is not None:
            doc = fget.__doc__
        self.__doc__ = doc

        self.kwargs = {}

    def __set_name__(self, owner, name):
        self.name = name

    def __call__(self, func):
        if self.fget is None:
            return self.getter(func)

        return self.setter(func)

    def __get__(self, instance, objtype=None):
        if instance is None:
            return self

        if self.fget is None:
            raise AttributeError('%s is a read-only property of %s' %
                                 (self.name, instance.__class__.__name__))

        return self.get(instance, objtype)

    def __set__(self, instance, value):

        if self.fset is None:
            raise AttributeError('%s is a write-only property of %s' %
                                 (self.name, instance.__class__.__name__))

        self.set(instance, value)

    def __delete__(self, instance):

        if self.fdel is None:
            raise AttributeError('%s is a permanent feat of %s' %
                                 (self.name, instance.__class__.__name__))

        self.delete(instance)

    @property
    def fget_signature(self):
        return inspect.signature(self.fget)

    def get(self, instance, objtype):
        return self.fget(instance)

    def set(self, instance, value):
        return self.fset(instance, value)

    def delete(self, instance):
        return self.fdel(instance)

    def getter(self, fget):
        return type(self)(fget, self.fset, self.fdel, self.__doc__,
                          **getattr(self, 'kwargs') or {})

    def setter(self, fset):
        return type(self)(self.fget, fset, self.fdel, self.__doc__,
                          **getattr(self, 'kwargs') or {})

    def deleter(self, fdel):
        return type(self)(self.fget, self.fset, fdel, self.__doc__,
                          **getattr(self, 'kwargs') or {})


class StorageProperty(NamedProperty):
    """A property that can store and retrieve information in the instance
    to which is attached.

    Requires that the owner class inherits StorageMixin.

    Notes
    -----

    The information is stored in uniquely a specified namespace defined by
    the derived class. Inside that storage, another namespace is specified
    using the property name.

    Derived class should use the dynamically created _store_get and _store_set
    to retrieve and store information.

    Derived classes must override the following variables:

    _storage_ns : str
        Defines a unique namespace under which the information of the
        derived class is stored.

    _storage_ns_init : callable
        Called upon initialization of the storage to initialize the
        specific storage of the namespace.

    """

    # Stores namespace to StorageProperty subclass
    # It cannot be dunder because it is accessed by __init_subclass__
    _storage_sub_ns_cls = weakref.WeakValueDictionary()

    _storage_ns = ''
    _storage_ns_init = None

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

        ns = cls._storage_ns
        if not ns:
            raise ValueError('Class %s must specify a storage namespace '
                             ' as required by StorageProperty' % cls)

        if ns in cls._storage_sub_ns_cls:
            if not issubclass(cls, cls._storage_sub_ns_cls[ns]):
                raise ValueError('Class %r storage namespace (%s) collides with '
                                 'class %r' % (cls, ns, cls._storage_sub_ns_cls[ns]))
        else:
            if cls._storage_ns_init is None:
                raise ValueError('Class %s must specify a storage initializer '
                                 ' as required by StorageProperty' % cls)

            cls._storage_sub_ns_cls[ns] = cls

            # Create partial versions of _store_get and _store_get
            #  with the corresponding namespace
            #  and store it in the specific subclass.
            cls._store_get = ft.partial(cls._ns_store_get, namespace=cls._storage_ns)
            cls._store_set = ft.partial(cls._ns_store_set, namespace=cls._storage_ns)
            cls._store_del = ft.partial(cls._ns_store_del, namespace=cls._storage_ns)

    def __set_name__(self, owner, name):
        require(self, owner, name, StorageMixin)

        super().__set_name__(owner, name)

    def _store_get(self, instance):
        return self._ns_store_get(instance, self._storage_ns)

    def _store_set(self, instance, value):
        return self._ns_store_set(instance, value, self._storage_ns)

    def _store_del(self, instance):
        return self._ns_store_del(instance, self._storage_ns)

    def _ns_store_get(self, instance, namespace):
        sto = instance.storage

        if namespace not in sto:
            cls = self._storage_sub_ns_cls[namespace]
            sto[namespace] = cls._storage_ns_init(instance)

        return sto[namespace][self.name]

    def _ns_store_set(self, instance, value, namespace):
        sto = instance.storage

        if namespace not in sto:
            cls = self._storage_sub_ns_cls[namespace]
            sto[namespace] = cls._storage_ns_init(instance)

        sto[namespace][self.name] = value

    def _ns_store_del(self, instance, namespace):
        sto = instance.storage

        if namespace not in sto:
            cls = self._storage_sub_ns_cls[namespace]
            sto[namespace] = cls._storage_ns_init(instance)

        try:
            del sto[namespace][self.name]
        except KeyError:
            pass


class StatsProperty(StorageProperty):
    """A property that keep stats on get and set calls.

    Stats can be retrieved with the `stat` methods and
    the following keys:
    - get
    - set
    - failed_get
    - failed_set

    The following statistics are provided in a namedtuple
    last : float
        most recent duration (seconds).
    count : int
        number of operations.
    mean : float
        average duration per operation (seconds).
    std : float
        standard deviation of the duration (seconds).
    min : float
        shortest duration (seconds).
    max : float
        longest duration (seconds).


    Requires that the owner class inherits StorageMixin.
    """

    _storage_ns = 'stats'
    _storage_ns_init = lambda _: defaultdict(RunningStats)

    def get(self, instance, objtype):

        with StatsProperty._store_get(self, instance).time('get'):
            return super().get(instance, objtype)

    def set(self, instance, value):

        with StatsProperty._store_get(self, instance).time('set'):
            return super().set(instance, value)

    def stats(self, instance, key):
        return StatsProperty._store_get(self, instance).stats(key)


class LogProperty(NamedProperty):
    """A property that log operations.

    Requires that the owner class inherits LogMixin.
    """

    log_values = True

    def __init__(self, *args, **kwargs):
        self.log_values = kwargs.pop('log_values', True)
        super().__init__(*args, **kwargs)

        self.kwargs.update({'log_values': self.log_values})

    def __set_name__(self, owner, name):
        require(self, owner, name, BaseLogMixin)

        super().__set_name__(owner, name)

    def _to_log(self, instance, value):
        if self.log_values is True:
            return value

        elif callable(self.log_values):
            try:
                return self.log_values(value)
            except Exception as e:
                instance.log_error('Could not convert value to log in %s, logging type: e', self.name, e)

        return type(value)

    def get(self, instance, objtype):

        instance.log_info('Getting %s', self.name)
        try:
            value = super().get(instance, objtype)
            log_value = self._to_log(instance, value)
            instance.log_debug('Got %s for %s', log_value, self.name)
        except Exception as e:
            instance.log_error('While getting %s: %s', self.name, e)
            raise e

        return value

    def set(self, instance, value):
        log_value = self._to_log(instance, value)
        instance.log_debug('Setting %s to %s', self.name, log_value)
        try:
            super().set(instance, value)
            instance.log_debug('%s was set to %s', self.name, log_value)
        except Exception as e:
            instance.log_error('While setting %s to %s: %s', self.name, log_value, e)
            raise e


class LockProperty(NamedProperty):
    """A property that with a set or get Lock.

    Requires that the owner class inherits LogMixin.
    """

    def __set_name__(self, owner, name):
        require(self, owner, name, LockMixin)

        super().__set_name__(owner, name)

    def get(self, instance, objtype):

        with instance.lock:
            return super().get(instance, objtype)

    def set(self, instance, value):
        with instance.lock:
            return super().set(instance, value)


class InstanceConfigurableProperty(StorageProperty):
    """A property that contains owner instance specific configuration variable.

    Requires that the owner class inherits StorageMixin.
    """

    _storage_ns = 'iconfig'
    _storage_ns_init = lambda _: defaultdict(dict)

    _config_keys = None
    _config = None
    _config_unset_value = None

    def __init__(self, *args, **kwargs):
        self._config = {}
        for k in self._config_keys:
            if k in kwargs:
                self._config[k] = kwargs.pop(k)

        super().__init__(*args, **kwargs)

        self.kwargs.update(self._config)

    def config_get(self, instance, key):

        if instance is None:
            return self._config.get(key, self._config_unset_value)
        try:
            return InstanceConfigurableProperty._store_get(self, instance)[key]
        except KeyError:
            return self._config.get(key, None)

    def config_set(self, instance, key, value):
        if instance is None:
            self._config[key] = value
        else:
            InstanceConfigurableProperty._store_get(self, instance)[key] = value

        self.on_config_set(instance, key, value)

    def config_iter(self, instance):
        for key in self._config_keys:
            yield key, self.config_get(instance, key)

    def on_config_set(self, instance, key, value):
        pass


class TransformProperty(InstanceConfigurableProperty):
    """A property that can transform value before a set operation
    or after a get operation.

    Requires that the owner class inherits InstanceConfigurableProperty.
    """

    pre_set = Config()
    post_get = Config()

    def __set_name__(self, owner, name):
        require(self, owner, name, StorageMixin, BaseLogMixin)

        super().__set_name__(owner, name)

    def get(self, instance, objtype):

        value = super().get(instance, objtype)

        transform = self.post_get_iget(instance)

        if not transform:
            return value

        try:
            value = transform(value)
            instance.log_debug('<T> Got %s for %s', value, self.name)
            return value
        except Exception as e:
            instance.log_error('While post-processing %s for %s: %s', value, self.name, e)
            raise e

    def set(self, instance, value):

        transform = self.pre_set_iget(instance)

        if transform:

            try:
                value = transform(value)
                instance.log_info('<T> Setting %s = %s', self.name, value)
            except Exception as e:
                instance.log_error('While pre-processing %s for %s: %s', value, self.name, e)
                raise e

        super().set(instance, value)


class CacheProperty(StorageProperty):
    """A property that can store, recall or invalidate a cache.
    """

    _storage_ns = 'cache'
    _storage_ns_init = lambda instance: missingdict(instance._cache_unset_value)

    def __set_name__(self, owner, name):
        require(self, owner, name, CacheMixin, BaseLogMixin)

        super().__set_name__(owner, name)

    def recall(self, instance):
        return CacheProperty._store_get(self, instance)

    def store(self, instance, value):
        CacheProperty._store_set(self, instance, value)

    def invalidate_cache(self, instance):
        CacheProperty._store_del(self, instance)


class GetCacheProperty(CacheProperty):
    """A property that stores the get value in the cache.

    Requires that the owner class inherits StorageMixin.
    """

    def get(self, instance, objtype):

        value = super().get(instance, objtype)

        self.store(instance, value)

        return value


class SetCacheProperty(CacheProperty):
    """A property that stores the set value in the cache.

    Requires that the owner class inherits StorageMixin.
    """

    def set(self, instance, value):
        super().set(instance, value)

        self.store(instance, value)


class GetSetCacheProperty(GetCacheProperty, SetCacheProperty):
    """A property that stores the get or set value in the cache.

    Requires that the owner class inherits StorageMixin.
    """


class PreventUnnecessarySetProperty(SetCacheProperty):
    """A property that prevents unnecessary set operations by comparing
    the value in the cache with the value to be set.

    Requires that the owner class inherits CacheMixin and LogMixin.
    """

    def set(self, instance, value):
        current_value = self.recall(instance)

        if value == current_value:
            instance.log_info('No need to set %s = %s (current=%s)', self.name, value, current_value)
            return

        super().set(instance, value)

    def force_set(self, instance, value):
        self.invalidate_cache(instance)
        self.set(instance, value)


class ReadOnceProperty(InstanceConfigurableProperty, GetCacheProperty):

    read_once = Config()

    def get(self, instance, owner=None):
        if self.read_once_iget(instance) and self.recall(instance) is not instance._cache_unset_value:
            return self.recall(instance)

        return super().get(instance, owner)


class ObservableProperty(CacheProperty):
    """A property that emits a signal when the cached value is changed
    (either via set or get)
    """

    def __set_name__(self, owner, name):
        if isinstance(name, str):
            require(self, owner, name, CacheMixin, ObservableMixin)
            setattr(owner, name + '_changed', owner._observer_signal_init())

        super().__set_name__(owner, name)

    def store(self, instance, value):
        if isinstance(self.name, DictPropertyNameKey):
            old_value = self.recall(instance)
            super().store(instance, value)
            getattr(instance, self.name.name + '_changed').emit(value, old_value, self.name.key)
        else:
            old_value = self.recall(instance)
            super().store(instance, value)
            getattr(instance, self.name + '_changed').emit(value, old_value)




from collections import UserDict, namedtuple


class DictPropertyNameKey(namedtuple('DictPropertyNameKey', 'name key')):

    def __str__(self):
        return '%s[%r]' % (self.name, self.key)


class missingdict(UserDict):
    """Dictionary that returns UNSET
    """

    __unset_value = None

    def __init__(self, unset_value, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__unset_value = unset_value

    def __missing__(self, key):
        return self.__unset_value


def keep_if_not(_skip=None, **kwargs):
    return {k: v for k, v in kwargs.items() if v is not _skip}


def require(inst, owner, name, *parents):
    """Raises a (nicely explained) type error if owner is not a subclass of all parents.

    It is useful when making sure that a certain type of property is only used within
    certain classes.

    Parameters
    ----------
    inst
        instance of a property object or method.
    owner : type
        class of an object.
    name : str
        name of the attribute to which the property object is assigned.
    parents : iterable of type
        classes that owner must inherit.

    Raises
    ------
    TypeError
        If owner does not inherit from all parents.

    """
    for parent in parents:
        if not issubclass(owner, parent):
            raise TypeError('%r is a %r but %r is not a subclass from %r as required' %
                            (name, inst.__class__.__name__, owner.__name__, parent.__name__))


def require_any(inst, owner, name, *parents):
    """Raises a (nicely explained) type error if owner is not a subclass of at least one parent.

    It is useful when making sure that a certain type of property is only used within
    certain classes.

    Parameters
    ----------
    inst
        instance of a property object.
    owner : type
        class of an object.
    name : str
        name of the attribute to which the property object is assigned.
    parents : iterable of type
        classes that owner must inherit at least one.

    Raises
    ------
    TypeError
        If owner does not inherit at least from one parents.

    """
    for parent in parents:
        if issubclass(owner, parent):
            break
    else:
        raise TypeError('%r is a %r but %r is not a subclass from any of %r as required' %
                        (name, inst.__class__.__name__, owner.__name__, parents))


class Config:

    def __set_name__(self, owner, name):
        from .props import InstanceConfigurableProperty
        from .methods import InstanceConfigurableMethod
        require_any(self, owner, name, InstanceConfigurableProperty, InstanceConfigurableMethod)
        if owner._config_keys is None:
            owner._config_keys = set()

        owner._config_keys.add(name)

        def _get(selfie):
            return selfie.config_get(None, name)

        def _set(selfie, value):
            return selfie.config_set(None, name, value)

        def _iget(selfie, instance):
            return selfie.config_get(instance, name)

        def _iset(selfie, instance, value):
            return selfie.config_set(instance, name, value)

        setattr(owner, name, property(_get, _set))
        setattr(owner, name + '_iget', _iget)
        setattr(owner, name + '_iset', _iset)









import unittest
import logging

from pimpmyclass import mixins, methods


class MemHandler(logging.Handler):



    def __init__(self):
        super().__init__()
        self.setFormatter(logging.Formatter(style='{'))
        self.history = list()

    def emit(self, record):
        self.history.append(self.format(record))


def define(proptype, *bases):

    class Dummy(*bases):

        logger_name = 'testing123'

        value = 3

        @proptype()
        def method(self):
            return self.value

        @proptype()
        def method2(self, n):
            return self.value * n

    return Dummy


class TestMethods(unittest.TestCase):

    def test_lock(self):

        with self.assertRaises(Exception):
            Dummy = define(methods.LockMethod)

        Dummy = define(methods.LockMethod, mixins.LockMixin)
        x = Dummy()
        self.assertEqual(x.method(), 3)

    def test_transformations(self):

        with self.assertRaises(Exception):
            Dummy = define(methods.TransformMethod)

        Dummy = define(methods.TransformMethod, mixins.StorageMixin, mixins.BaseLogMixin)
        x = Dummy()
        self.assertEqual(x.method(), 3)

        self.assertEqual(x.method2(2), 6)

        Dummy.method2.params = {'n': lambda x: 2*x}
        self.assertEqual(x.method2(2), 12)

    def test_log(self):

        with self.assertRaises(Exception):
            define(methods.LogMethod)

        Dummy = define(methods.LogMethod, mixins.LogMixin)
        x = Dummy()

        hdl = MemHandler()
        x.logger.addHandler(hdl)
        x.logger.setLevel(logging.DEBUG)

        x.method()
        x.method2(3)

        self.assertEqual(hdl.history, ['Calling method',
                                       'method returned 3',
                                       'Calling method2 with ((3,), {}))',
                                       'method2 returned 9'])

    def test_log_config_false(self):

        Dummy = define(lambda: methods.LogMethod(log_values=False), mixins.LogMixin)
        x = Dummy()

        hdl = MemHandler()
        x.logger.addHandler(hdl)
        x.logger.setLevel(logging.DEBUG)

        x.method()
        x.method2(3)

        self.assertEqual(hdl.history, ['Calling method',
                                       "method returned <class 'int'>",
                                       "Calling method2 with ((<class 'int'>,), {}))",
                                       "method2 returned <class 'int'>"])


    def test_log_config_fun(self):

        Dummy = define(lambda: methods.LogMethod(log_values=lambda x: '%s %s' % (x, type(x))), mixins.LogMixin)
        x = Dummy()

        hdl = MemHandler()
        x.logger.addHandler(hdl)
        x.logger.setLevel(logging.DEBUG)

        x.method()
        x.method2(3)

        self.assertEqual(hdl.history, ['Calling method',
                                       "method returned 3 <class 'int'>",
                                       """Calling method2 with (("3 <class 'int'>",), {}))""",
                                       "method2 returned 9 <class 'int'>"])
